# Division Algorithm

# Given a >= 0, b > 0 compute
#       a = b * q + r, 0 <= r < b
#   q = a div b
#   r = a rem b

# Based on the binary representation algorithm which in essence
# computes a div 1. The bin rep algorithm reduces the remainder by the
# largest possible power of 2 at each step. Instead of reducing by
# 2^i, reducing by b * 2^1 extracts out 2^i multiples of b.  The resulting
# sequence of bits is the representation of q, and the remaining value
# that is smaller than 2^0 * b is r.

# idiom to initialize PC with START addr
_:START;

START: ;

# bits posn    0   1 ... (nbits-1)
# bits value 2^0 2^1 ... 2^(nbits-1)

# Maximum number of bits - note this is an asm symbol not a mem cell
nbits = 8;

# Read in a and b
SUB, IN, _0, a;
SUB, IN, _0, b;

# Make sure a is >= 0, b is > 0
BGE, a, _0, aok;
SUB, _0, a, a;
aok:
BGE, b, _1, bok;
SUB, _0, b, b;
BGE, _0, b, error;
bok:

# Begin by computing a table of the multiples of b scaled by powers of 
# 2 associated.

# Pointer p_b2n to bit 0, incremented on each bit
SUB, b2n_start_addr, _0, p_b2n;

# Current value of b * 2^1
SUB, b, _0, two_to_i;

# loop i = 0 .. n-1 times, using i as the index
SUB, _0, _0, i;
loop_1_begin: BGE, i, n, loop_1_end;

    # indirect store:  [two_to_i] -> [ [p_b2n] ]
    # build instruction to store b2n into that address
    SUB, p_b2n, _0, @+7;
    SUB, two_to_i, _0, 0;

    # double the power of 2
    SUB, _0, two_to_i, tmp;
    SUB, two_to_i, tmp, two_to_i;

    # increment i and address of next b2n table entry
    SUB, i, _n1, i;
    SUB, p_b2n, _n1, p_b2n;

    # back to top of loop
    BGE, _0, _0, loop_1_begin;
loop_1_end:

# Initialize the remainder
SUB, a, _0, remain;

# end address is 1 beyond actual table, and we want to start
# at last entries and work down

SUB, b2n_end_addr, _1, p_b2n;
SUB, bitsq_end_addr, _1, p_bitsq;

# Work in decreasing powers of 2
# loop i = 0 .. n-1 times, using i as the index

SUB, _0, _0, i;

loop_2_begin: BGE, i, n, loop_2_end;

    # indirect load:  [ [p_b2n] ] -> [ two_to_i ]
    # build instruction to load 2^(n-1-i) into two_to_i
    SUB, p_b2n, _0, @+5;
    SUB, 0, _0, two_to_i;

    # remain < pow => remain does not contain that power so bit is 0
    # otw bit is 1

    SUB, _0, _0, bit;
    BGE, remain, two_to_i, bigger;
    BGE, _0, _0, smaller;
    bigger: 
        SUB, _1, _0, bit;
        # remove that power from remainder
        SUB, remain, two_to_i, remain;
    smaller:

    # build instruction to store the bit into the bitsq table.
    SUB, p_bitsq, _0, @+7;
    SUB, bit, _0, 0;

    # increment i and decrement address of current entry in power
    # table and bit value
    SUB, i, _n1, i;
    SUB, p_b2n, _1, p_b2n;
    SUB, p_bitsq, _1, p_bitsq;

    # back to top of loop
    BGE, _0, _0, loop_2_begin;
loop_2_end:

SUB, remain, _0, r;

# Now reconstruct q from its bits
SUB, bitsq_end_addr, _1, p_bitsq;
SUB, _0, _0, i;
SUB, _0, _0, q;

loop_3_begin: BGE, i, n, loop_3_end;
    # we want to compute q <- 2 * q + bits[n-i]

    # tmp <- 0 - bits[n-i] - q - q
    # q <- 0 - tmp

    # compute -2q, i.e.shift left 1 bit
    SUB, _0, q, tmp;
    SUB, tmp, q, tmp;

    # build instruction to subtract bit[n-i] from tmp
    SUB, p_bitsq, _0, @+6;
    SUB, tmp, 0, tmp;

    # and then q <- -tmp
    SUB, _0, tmp, q;

    # increment i and decrement address of current entry in power
    # table and bit value
    SUB, i, _n1, i;
    SUB, p_bitsq, _1, p_bitsq;

    # back to top of loop
    BGE, _0, _0, loop_3_begin;
loop_3_end:

SUB, a, _0, OUT;
SUB, b, _0, OUT;
SUB, q, _0, OUT;
SUB, r, _0, OUT;

# Keep asking for input until get error
BGE, _0, _0, START;

# Halt by setting PC to 0
SUB, _0, _0, PC;

# Error
error: SUB, _0, _0, PC;

# Some constants.  Note that if you put them in this order
# then you can index off _0, for example, this stores 1 - (-1) 
# into tmp.
#   SUB, _0+1, _0-1, tmp;

_n2: -2;
_n1: -1;
_0: 0;
_1: 1;
_2: 2;

# Inputs a and b
a: 0;
b: 0;
# Results q and r
q: 0;
r: 0;

# Current power 2^n
two_to_i: 1;

# Remainder bits to be reduced, should be < b when finished.
remain: 0;

# Current bit value
bit: 0;

# Variables

# Number of bits expected
n: nbits;

# Index
i: 0;

# Pointer to b2n entry
p_b2n: 0;

# Pointer to bitsq entry
p_bitsq: 0;

# Volatile tmp value
tmp: 0;

# the idiom to reserve space for a table of size k is
#
# table_start_addr: table - addr of element 0
# table_end_addr: table_end - 1 past element k
#
# table: ;
# @ = @ + size;
# table_end: ;

# power table is [b2n_start_addr ... b2n_end_addr] inclusive
b2n_start_addr: b2n;
b2n_end_addr: b2n_end;

# bitsq table is [bitsq_start_addr ... bitsq_end_addr - 1] 
bitsq_start_addr: bitsq;
bitsq_end_addr: bitsq_end;

# b * 2^i table starts here, b2n[i] = b * 2^i 0 <= i < nbits
b2n: ;
@=@+nbits;
b2n_end:;

# bits of binary rep of q start here, bitsq[i] = value of bit i,
# in little endian ordering. 
#   bitq[i] is value of bit representing b * 2^i

bitsq: ;
@=@+nbits;
bitsq_end: ;
